@@ -9,12 +9,20 @@ module Test.Hspec.Core.Example (
 , Result (..)
 , Location (..)
 , LocationAccuracy (..)
+, FailureReason (..)
+, safeEvaluateExample
 ) where
 
 import           Data.Maybe (fromMaybe)
 import           Data.List (isPrefixOf)
 import qualified Test.HUnit.Lang as HUnit
+
+#if MIN_VERSION_HUnit(1,4,0)
+import           Data.CallStack
+#endif
+
 import qualified Control.Exception as E
+import           Control.DeepSeq
 import           Data.Typeable (Typeable)
 import qualified Test.QuickCheck as QC
 import           Test.Hspec.Expectations (Expectation)
@@ -24,14 +32,12 @@ import qualified Test.QuickCheck.Property as QCP
 
 import           Test.Hspec.Core.QuickCheckUtil
 import           Test.Hspec.Core.Util
-import           Test.Hspec.Compat
+import           Test.Hspec.Core.Compat
 
 -- | A type class for examples
 class Example e where
   type Arg e
-#if __GLASGOW_HASKELL__ >= 704
   type Arg e = ()
-#endif
   evaluateExample :: e -> Params -> (ActionWith (Arg e) -> IO ()) -> ProgressCallback -> IO Result
 
 data Params = Params {
@@ -52,9 +58,18 @@ type ProgressCallback = Progress -> IO ()
 type ActionWith a = a -> IO ()
 
 -- | The result of running an example
-data Result = Success | Pending (Maybe String) | Fail (Maybe Location) String
+data Result = Success | Pending (Maybe String) | Failure (Maybe Location) FailureReason
   deriving (Eq, Show, Read, Typeable)
 
+data FailureReason = NoReason | Reason String | ExpectedButGot (Maybe String) String String
+    deriving (Eq, Show, Read, Typeable)
+
+instance NFData FailureReason where
+  rnf reason = case reason of
+    NoReason -> ()
+    Reason r -> r `deepseq` ()
+    ExpectedButGot p e a  -> p `deepseq` e `deepseq` a `deepseq` ()
+
 instance E.Exception Result
 
 -- | @Location@ is used to represent source locations.
@@ -74,9 +89,45 @@ data LocationAccuracy =
   BestEffort
   deriving (Eq, Show, Read)
 
+safeEvaluateExample :: Example e => e -> Params -> (ActionWith (Arg e) -> IO ()) -> ProgressCallback -> IO (Either E.SomeException Result)
+safeEvaluateExample example params around progress = do
+  r <- safeTry $ forceResult <$> evaluateExample example params around progress
+  return $ case r of
+    Left e | Just result <- E.fromException e -> Right result
+    Left e | Just hunit <- E.fromException e -> Right (hunitFailureToResult hunit)
+    _ -> r
+  where
+    forceResult :: Result -> Result
+    forceResult r = case r of
+      Success -> r
+      Pending m -> m `deepseq` r
+      Failure _ m -> m `deepseq` r
+
+instance Example Result where
+  type Arg Result = ()
+  evaluateExample e = evaluateExample (\() -> e)
+
+instance Example (a -> Result) where
+  type Arg (a -> Result) = a
+  evaluateExample example _params action _callback = do
+    ref <- newIORef Success
+    action (writeIORef ref . example)
+    readIORef ref
+
 instance Example Bool where
   type Arg Bool = ()
-  evaluateExample b _ _ _ = if b then return Success else return (Fail Nothing "")
+  evaluateExample e = evaluateExample (\() -> e)
+
+instance Example (a -> Bool) where
+  type Arg (a -> Bool) = a
+  evaluateExample p _params action _callback = do
+    ref <- newIORef Success
+    action $ \a -> example a >>= writeIORef ref
+    readIORef ref
+    where
+      example a
+        | p a = return Success
+        | otherwise = return (Failure Nothing NoReason)
 
 instance Example Expectation where
   type Arg Expectation = ()
@@ -85,25 +136,29 @@ instance Example Expectation where
 hunitFailureToResult :: HUnit.HUnitFailure -> Result
 hunitFailureToResult e = case e of
 #if MIN_VERSION_HUnit(1,3,0)
-  HUnit.HUnitFailure mLoc err -> Fail location err
+  HUnit.HUnitFailure mLoc err ->
+#if MIN_VERSION_HUnit(1,5,0)
+      case err of
+        HUnit.Reason reason -> Failure location (Reason reason)
+        HUnit.ExpectedButGot preface expected actual -> Failure location (ExpectedButGot preface expected actual)
+#else
+      Failure location (Reason err)
+#endif
     where
       location = case mLoc of
         Nothing -> Nothing
-        Just loc -> Just $ Location (HUnit.locationFile loc) (HUnit.locationLine loc) (HUnit.locationColumn loc) ExactLocation
+#if MIN_VERSION_HUnit(1,4,0)
+        Just loc -> Just $ Location (srcLocFile loc) (srcLocStartLine loc) (srcLocStartCol loc) ExactLocation
 #else
-  HUnit.HUnitFailure err -> Fail Nothing err
+        Just loc -> Just $ Location (HUnit.locationFile loc) (HUnit.locationLine loc) (HUnit.locationColumn loc) ExactLocation
+#endif
+#else
+  HUnit.HUnitFailure err -> Failure Nothing (Reason err)
 #endif
 
 instance Example (a -> Expectation) where
   type Arg (a -> Expectation) = a
-  evaluateExample e _ action _ = (action e >> return Success) `E.catches` [
-      E.Handler (return . hunitFailureToResult)
-    , E.Handler (return :: Result -> IO Result)
-    ]
-
-instance Example Result where
-  type Arg Result = ()
-  evaluateExample r _ _ _ = return r
+  evaluateExample e _ action _ = action e >> return Success
 
 instance Example QC.Property where
   type Arg QC.Property = ()
@@ -116,11 +171,11 @@ instance Example (a -> QC.Property) where
     return $
       case r of
         QC.Success {}               -> Success
-        QC.Failure {QC.output = m}  -> fromMaybe (Fail Nothing $ sanitizeFailureMessage r) (parsePending m)
-        QC.GaveUp {QC.numTests = n} -> Fail Nothing ("Gave up after " ++ pluralize n "test" )
-        QC.NoExpectedFailure {}     -> Fail Nothing ("No expected failure")
+        QC.Failure {QC.output = m}  -> fromMaybe (Failure Nothing . Reason $ sanitizeFailureMessage r) (parsePending m)
+        QC.GaveUp {QC.numTests = n} -> Failure Nothing (Reason $ "Gave up after " ++ pluralize n "test" )
+        QC.NoExpectedFailure {}     -> Failure Nothing (Reason $ "No expected failure")
 #if MIN_VERSION_QuickCheck(2,8,0)
-        QC.InsufficientCoverage {}  -> Fail Nothing ("Insufficient coverage")
+        QC.InsufficientCoverage {}  -> Failure Nothing (Reason $ "Insufficient coverage")
 #endif
     where
       qcProgressCallback = QCP.PostTest QCP.NotCounterexample $
